@@ -10,7 +10,7 @@ import numpy as np
 import xarray as xr
 import pandas as pd
 import tethys_utils as tu
-from tethysts.utils import s3_client, get_object_s3, read_json_zstd, b2_public_key_pattern, read_pkl_zstd
+from tethysts.utils import s3_client, get_object_s3, read_json_zstd, read_pkl_zstd, download_results, create_public_s3_url
 import tethys_data_models as tdm
 from typing import List, Optional, Dict, Union
 # import boto3
@@ -33,6 +33,7 @@ from datetime import date, datetime
 import multiprocessing as mp
 import pyproj
 import random
+from tethysts import Tethys
 
 ##############################################
 ### Parameters
@@ -41,7 +42,7 @@ b2_dl_by_id = 'https://b2.tethys-ts.xyz/b2api/v1/b2_download_file_by_id?fileId={
 
 
 preprocessed_dir = 'preprocessed_data'
-previous_dir = 'previous_results'
+previous_dir = 'previous_data'
 final_dir = 'final_results'
 
 preprocessed_file_str = '{ds_id}_{file_id}.nc.zst'
@@ -55,6 +56,10 @@ base_encoding = {'lon': {'dtype': 'int32', '_FillValue': -999999, 'scale_factor'
 
 dask.config.set(**{'array.slicing.split_large_chunks': False})
 
+agg_stat_mapping = {'mean': 'mean', 'cumulative': 'sum', 'continuous': None, 'maximum': 'max', 'median': 'median', 'minimum': 'min', 'mode': 'mode', 'sporadic': None, 'standard_deviation': 'std', 'incremental': 'cumsum'}
+
+tz_str = 'Etc/GMT{0:+}'
+
 ###############################################
 ### Helper functions
 
@@ -191,17 +196,30 @@ def bulk_estimate_time_interval(data, block_lengths, null_grid=None, max_mb=2):
 
 def estimate_time_interval_rough(data_grid_length, time_freq, block_lengths, max_mb=2, values_per_mb=550000):
     """
+    Function to roughly estimate the appropriate time intervals based on specific block lengths.
 
+    Parameters
+    ----------
+    data_grid_length: float or int
+        The grid resolution.
+    time_freq: str
+        The time resolution in pandas time freq format.
+    block_lengths: list of float
+        The block lengths to test. Should be equal to or greater than the data_grid_length.
+    max_mb: float or int
+        The max size of the results object.
+    values_per_mb: int
+        The number of data values in a results object per MB. Do not change unless you've done the testing to determine this value.
     """
     t1 = pd.date_range('2000-01-01', '2000-01-08', freq = time_freq)[:-1]
     val_per_day = len(t1)/7
 
-    total_val_per_file = max_mb * values_per_mb
+    total_val_per_day = max_mb * values_per_mb
 
     res_list = []
     for bl in block_lengths:
         n_stns = int(np.ceil((bl**2)/(data_grid_length**2)))
-        days_per_chunk = int(total_val_per_file/n_stns/val_per_day)
+        days_per_chunk = int(total_val_per_day/n_stns/val_per_day)
         dict1 = {'time_interval': days_per_chunk,
                  'n_geometries_per_station': n_stns}
         res_list.append(dict1)
@@ -227,13 +245,18 @@ def decompress_path(glob_path, compression_type='gzip', max_workers=4, **kwargs)
 def file_format_conversion(glob_path, file_format='grib', max_workers=4, **kwargs):
     """
     Function to convert data files to netcdf files.
-    This also should add in the time coordinate when converting TIF files. It should parse the time from the file name.
     """
     if isinstance(glob_path, str):
         files1 = glob.glob(glob_path)
     elif isinstance(glob_path, list):
         files1 = glob_path
-    new_files = tu.data_io.convert_gribs_remove(files1, max_workers=max_workers)
+
+    if file_format == 'grib':
+        new_files = tu.data_io.convert_gribs_remove(files1, max_workers=max_workers)
+    else:
+        raise NotImplementedError('file_format not available.')
+    # elif file_format == 'geotiff':
+    #     new_files = tu.data_io.convert_geotiffs_to_nc(files1, max_workers=max_workers, **kwargs)
 
     return new_files
 
@@ -476,7 +499,7 @@ def combine_metadata(project, dataset):
     return datasets
 
 
-def multi_calc_new_variables(nc_paths, datasets, func_dict):
+def multi_calc_new_variables(nc_paths, datasets, version_date, func_dict):
     """
 
     """
@@ -488,7 +511,7 @@ def multi_calc_new_variables(nc_paths, datasets, func_dict):
     new_paths_list = []
     for i, g in dates1.groupby('date'):
         # print(g)
-        p = tu.processing.calc_new_variables(g['path'].tolist(), datasets, func_dict)
+        p = tu.processing.calc_new_variables(g['path'].tolist(), datasets, version_date, func_dict)
         new_paths_list.append(p)
 
     # elif max_workers > 1:
@@ -550,11 +573,6 @@ def multi_mergetime_nc_remove(source_paths, by, max_workers=3):
     return new_paths
 
 
-
-
-
-
-
 ###############################################
 ### Class
 
@@ -576,14 +594,14 @@ class Titan(object):
 
         ## Add in the additional paths
         preprocessed_path = os.path.join(temp_path, preprocessed_dir)
-        # previous_path = os.path.join(temp_path, previous_dir)
+        previous_path = os.path.join(temp_path, previous_dir)
         # final_path = os.path.join(temp_path, final_dir)
         os.makedirs(preprocessed_path, exist_ok=True)
-        # os.makedirs(previous_path, exist_ok=True)
+        os.makedirs(previous_path, exist_ok=True)
         # os.makedirs(final_path, exist_ok=True)
 
         self.preprocessed_path = preprocessed_path
-        # self.previous_path = previous_path
+        self.previous_path = previous_path
         # self.final_path = final_path
 
         ## Test for the run_id and key
@@ -638,12 +656,12 @@ class Titan(object):
 
 
     @staticmethod
-    def process_sparse_stations_from_df(stns, dataset_id=None, connection_config=None, bucket=None, system_version=4):
+    def process_sparse_stations_from_df(stns, precision=7):
         """
         Function that takes a stns dataframe of station data and converts it to an Xarray Dataset for Tethys. This is ultimately meant to be combined with the time series data for futher processing. If a geometry column is provided, it must be as a geojson-type dict (not a geopandas column).
 
         """
-        stns = tu.processing.process_sparse_stations_from_df(stns, dataset_id, connection_config=connection_config, bucket=bucket, system_version=system_version)
+        stns = tu.processing.process_sparse_stations_from_df(stns, precision=precision)
 
         return stns
 
@@ -662,6 +680,102 @@ class Titan(object):
         return datasets
 
 
+    @staticmethod
+    def combine_obs_stn_data(ts_data, stn_data, mod_date=False):
+        """
+        Function to take a time series DataFrame and station data (in 3 formats) and combine them into a single xr.Dataset.
+
+        Parameters
+        ----------
+        ts_data: pd.DataFrame
+            The DataFrame should have height and time as columns in addition to the parameter column.
+        stn_data: pd.Series, pd.DataFrame, dict, xr.Dataset
+            The station data that should have geometry as a column.
+        mod_date: bool
+            The the modified_date be added to the ts_data?
+
+        Returns
+        -------
+        xr.Dataset
+        """
+        r1 = tu.processing.combine_obs_stn_data(ts_data, stn_data, mod_date=mod_date)
+
+        return r1
+
+
+    def resample_time_series(self, df, dataset_id, sum_closed='right', other_closed='left', discrete=False):
+        """
+
+        """
+        dataset = self.datasets[dataset_id]
+        freq_code = dataset['frequency_interval']
+        agg_stat = dataset['aggregation_statistic']
+        parameter = dataset['parameter']
+        utc_offset = dataset['utc_offset']
+
+        df1 = df.copy()
+
+        # check columns
+        cols = df1.columns
+
+        if 'time' not in cols:
+            raise ValueError('time must be a column.')
+        if parameter not in cols:
+            raise ValueError(parameter + ' must be a column.')
+
+        grp = []
+        if 'station_id' in cols:
+            grp.append('station_id')
+        if 'height' in cols:
+            grp.append('height')
+
+        vars1 = [parameter] + ['time'] + grp
+
+        ancillary_variables = [v for v in cols if (v not in vars1)]
+
+        # main_vars = [parameter] + ancillary_variables
+
+        # Convert times to local TZ if necessary
+        if (not freq_code in ['T', 'H', '1H']) and (not utc_offset == '0H'):
+            t1 = int(utc_offset.split('H')[0])
+            tz1 = tz_str.format(-t1)
+            df1['time'] = df1['time'].dt.tz_localize('UTC').dt.tz_convert(tz1).dt.tz_localize(None)
+
+        ## Aggregate data if necessary
+        # Parameter
+        if freq_code == 'T':
+            data1 = df1
+        else:
+            agg_fun = agg_stat_mapping[agg_stat]
+
+            if agg_fun == 'sum':
+                data0 = tu.misc.grp_ts_agg(df1[vars1], grp, 'time', freq_code, agg_fun, closed=sum_closed)
+            else:
+                data0 = tu.misc.grp_ts_agg(df1[vars1], grp, 'time', freq_code, agg_fun, discrete, closed=other_closed)
+
+            # Ancillary variables
+            av_list = [data0]
+            for av in ancillary_variables:
+                if 'quality_code' == av:
+                    df1['quality_code'] = pd.to_numeric(df1['quality_code'], errors='coerce', downcast='integer')
+                    qual1 = tu.misc.grp_ts_agg(df1[['time'] + grp + ['quality_code']], grp, 'time', freq_code, 'min')
+                    av_list.append(qual1)
+                else:
+                    av1 = tu.misc.grp_ts_agg(df1[['time'] + grp + [av]], grp, 'time', freq_code, 'max')
+                    av_list.append(av1)
+
+            # Put the data together
+            data1 = pd.concat(av_list, axis=1).reset_index()
+
+        # Convert time back to UTC if necessary
+        if (not freq_code in ['T', 'H', '1H']) and (not utc_offset == '0H'):
+            data1['time'] = data1['time'].dt.tz_localize(tz1).dt.tz_convert('utc').dt.tz_localize(None)
+
+        return data1
+
+
+
+
     # def init_results_dict(self):
     #     """
 
@@ -726,7 +840,7 @@ class Titan(object):
             if exist_version != system_version:
                 print('The bucket already has Tethys data and is version {}. You have been warned...'.format(exist_version))
 
-        print('All checks passed!')
+        print('All status checks passed!')
 
         ## Save parameters
         setattr(self, 'connection_config', connection_config)
@@ -776,7 +890,7 @@ class Titan(object):
         self.diagnostics['diagnostic_s3_obj_id'] = obj_resp['VersionId']
 
         if isinstance(self.public_url, str):
-            dgn_url = b2_public_key_pattern.format(base_url=self.public_url, bucket=self.bucket, obj_key=dgn_key)
+            dgn_url = create_public_s3_url(self.public_url, self.bucket, dgn_key)
         else:
             dgn_url = ''
 
@@ -793,55 +907,65 @@ class Titan(object):
 
         # TODO: Check for existing datasets in other S3 buckets
 
+        ## Check if the chunking parameters are in the datasets
+        for ds in dataset_list:
+            if 'chunk_parameters' in ds:
+                _ = tdm.dataset.ChunkParams(**ds['chunk_parameters'])
+            elif ds['method'] in ['sensor_recording', 'field_activity', 'sample_analysis']:
+                print('Default chunk_parameters have been set.')
+                ds['chunk_parameters'] = {'block_length': 0, 'time_interval': 7300}
+            else:
+                raise ValueError('chunk_parameters have not been set in the dataset metadata. Please do so.')
+
+        ## Might keep this for later...
+        # chunk_params_bool_list = []
+        # for ds in dataset_list:
+        #     if 'chunk_parameters' in ds:
+        #         _ = tdm.dataset.ChunkParams(**ds['chunk_parameters'])
+        #         chunk_params_bool_list.extend([True])
+        #     else:
+        #         chunk_params_bool_list.extend([True])
+
+        # if all(chunk_params_bool_list):
+        #     self.diagnostics['set_chunking_parameters'] = {'pass': True}
+
+        ## Validate dataset model
+        for ds in dataset_list:
+            _ = tdm.dataset.Dataset(**ds)
+
+        ## Set attributes
         ds_dict = {ds['dataset_id']: ds for ds in dataset_list}
 
         setattr(self, 'dataset_list', dataset_list)
         setattr(self, 'datasets', ds_dict)
 
-        ## Check if the chunking parameters are in the datasets
-        chunk_params_bool_list = []
-        for ds in dataset_list:
-            if 'chunk_parameters' in ds:
-                _ = tdm.dataset.ChunkParams(**ds['chunk_parameters'])
-                chunk_params_bool_list.extend([True])
-            else:
-                chunk_params_bool_list.extend([True])
-
-        if all(chunk_params_bool_list):
-            self.diagnostics['set_chunking_parameters'] = {'pass': True}
-
         ## diagnostic log
         self.diagnostics['load_dataset_metadata'] = {'pass': True}
         self.diagnostics['attributes'].update({'dataset_list': dataset_list, 'datasets': ds_dict})
 
 
-    def set_chunking_parameters(self, geometry_count=1, time_interval=None, height_interval=None):
+    # def set_chunking_parameters(self, block_length, time_interval=None):
+    #     """
+
+    #     """
+    #     tu.misc.diagnostic_check(self.diagnostics, 'load_dataset_metadata')
+
+    #     chunk_params = tdm.dataset.ChunkParams(block_length=block_length, time_interval=time_interval).dict(exclude_none=True)
+
+    #     datasets = []
+    #     for ds in self.dataset_list:
+    #         ds['chunk_parameters'] = chunk_params
+    #         datasets.append(ds)
+
+    #     self.load_dataset_metadata(datasets)
+
+
+    def process_versions(self, version_data: dict):
         """
 
         """
         tu.misc.diagnostic_check(self.diagnostics, 'load_dataset_metadata')
 
-        chunk_params = tdm.dataset.ChunkParams(geometry_count=geometry_count, time_interval=time_interval, height_interval=height_interval).dict(exclude_none=True)
-
-        datasets = []
-        for ds in self.dataset_list:
-            ds['chunk_parameters'] = chunk_params
-            datasets.append(ds)
-
-        self.load_dataset_metadata(datasets)
-
-
-    def process_versions(self, version_data: dict = None):
-        """
-
-        """
-        tu.misc.diagnostic_check(self.diagnostics, 'load_dataset_metadata')
-
-        if isinstance(self.public_url, str):
-            conn_config = self.public_url
-        else:
-            conn_config = self.connection_config
-
         version_dict = tu.s3.process_results_versions(self.dataset_list, bucket=self.bucket, connection_config=self.connection_config, public_url=self.public_url, version_data=version_data, system_version=self.system_version)
         max_version_date_key = tu.misc.make_run_date_key(max(list([d['version_date'] for d in version_dict.values()])))
 
@@ -856,7 +980,7 @@ class Titan(object):
         return version_dict
 
 
-    def save_preprocessed_results(self, dataset_id: str, data: Union[List[xr.Dataset], xr.Dataset]):
+    def save_preprocessed_results(self, data: Union[List[xr.Dataset], xr.Dataset], dataset_id: str):
         """
         Method to save preprocessed results xarray datasets to netcdf files in the temp_path. The results must have the standard dimensions, geometry caclulated, appropriate parameter names, and station data.
         """
@@ -887,9 +1011,19 @@ class Titan(object):
         ## Save the data
         file_list = []
         for d in data_list:
+            # print(str(d.ref.values[0]))
+
+            ## Update the metadata
+            d = tu.processing.add_metadata_results(d, dataset1, self.max_version_date_key)
+
             file_id = uuid.uuid4().hex[:14]
             file_path = os.path.join(self.preprocessed_path, preprocessed_file_str.format(ds_id=dataset_id, file_id=file_id))
+            # d.to_netcdf(file_path)
             _ = tu.misc.write_pkl_zstd(d.to_netcdf(), file_path)
+
+            # Read the file back in to make sure it worked...
+            # _ = xr.load_dataset(read_pkl_zstd(file_path))
+
             file_list.append(file_path)
 
         return file_list
@@ -933,11 +1067,11 @@ class Titan(object):
     #     return new_paths1
 
 
-    def file_format_conversion(self, source_paths, file_format='grib', max_workers=4):
+    def file_format_conversion(self, source_paths, file_format='grib', max_workers=4, **kwargs):
         """
 
         """
-        new_paths2 = file_format_conversion(source_paths, file_format=file_format, max_workers=max_workers)
+        new_paths2 = file_format_conversion(source_paths, file_format=file_format, max_workers=max_workers, **kwargs)
 
         return new_paths2
 
@@ -964,7 +1098,7 @@ class Titan(object):
         """
 
         """
-        new_paths2 = multi_calc_new_variables(source_paths, self.dataset_list, func_dict)
+        new_paths2 = multi_calc_new_variables(source_paths, self.dataset_list, self.max_version_date_key, func_dict)
 
         return new_paths2
 
@@ -973,8 +1107,8 @@ class Titan(object):
         """
 
         """
-        if by == 'station':
-            tu.misc.diagnostic_check(self.diagnostics, 'copy_interim_objs')
+        # if by == 'station':
+        #     tu.misc.diagnostic_check(self.diagnostics, 'copy_interim_objs')
 
         new_paths = multi_mergetime_nc_remove(source_paths, by=by, max_workers=max_workers)
 
@@ -1087,11 +1221,11 @@ class Titan(object):
         return new_paths
 
 
-    def save_final_results(self, source_paths, correct_times=False, max_workers=4):
+    def save_new_results(self, source_paths, correct_times=False, max_workers=4):
         """
 
         """
-        tu.misc.diagnostic_check(self.diagnostics, 'merge_nc_files')
+        # tu.misc.diagnostic_check(self.diagnostics, 'merge_nc_files')
         tu.misc.diagnostic_check(self.diagnostics, 'process_versions')
 
         ## Put the big files first
@@ -1101,40 +1235,181 @@ class Titan(object):
         source_paths1 = merge_df.sort_values('size', ascending=False).path.tolist()
 
         ## Iterate through files
-        with concurrent.futures.ProcessPoolExecutor(max_workers=max_workers, mp_context=mp.get_context("spawn")) as executor:
-            futures = []
+        if max_workers <= 1:
+            results_new_paths1 = []
             for source_path in source_paths1:
                 ds_id = source_path.split('/')[-1].split('_')[0]
                 metadata = self.datasets[ds_id]
-                f = executor.submit(tu.processing.save_final_results, source_path, metadata, self.max_version_date_key, correct_times, self.system_version)
-                futures.append(f)
-            runs = concurrent.futures.wait(futures)
+                f = tu.processing.save_new_results(source_path, metadata, self.max_version_date_key, correct_times, self.system_version)
+                results_new_paths1.extend(f)
+        else:
+            with concurrent.futures.ProcessPoolExecutor(max_workers=max_workers, mp_context=mp.get_context("spawn")) as executor:
+                futures = []
+                for source_path in source_paths1:
+                    ds_id = source_path.split('/')[-1].split('_')[0]
+                    metadata = self.datasets[ds_id]
+                    f = executor.submit(tu.processing.save_new_results, source_path, metadata, self.max_version_date_key, correct_times, self.system_version)
+                    futures.append(f)
+                runs = concurrent.futures.wait(futures)
 
-        ## process output
-        results_new_paths1 = []
-        stn_new_paths1 = []
-
-        new_paths = [r.result() for r in runs[0]]
-
-        for paths_list in new_paths:
-            r1, s1 = paths_list
-            results_new_paths1.extend(r1)
-            stn_new_paths1.extend(s1)
+            ## process output
+            results_new_paths1 = []
+            extend = results_new_paths1.extend
+            _ = [extend(r.result()) for r in runs[0]]
 
         results_new_paths1.sort()
-        stn_new_paths1.sort()
 
         ## diagnostic log
-        self.diagnostics['save_final_results'] = {'pass': True}
+        self.diagnostics['save_new_results'] = {'pass': True}
 
-        return results_new_paths1, stn_new_paths1
+        return results_new_paths1
 
 
-    def upload_final_results(self, results_paths, threads=30):
+    def update_conflicting_results(self, results_paths, threads=60, max_workers=4):
         """
 
         """
-        tu.misc.diagnostic_check(self.diagnostics, 'save_final_results')
+        ## Determine which chunks need to uploaded
+        if isinstance(self.public_url, str):
+            remote1 = {'bucket': self.bucket, 'public_url': self.public_url, 'version': self.system_version}
+        else:
+            remote1 = {'bucket': self.bucket, 'connection_config': self.connection_config, 'version': self.system_version}
+
+        r_chunks = tu.s3.determine_results_chunks_diffs(results_paths, remote1)
+
+        r_chunks1 = r_chunks[r_chunks['_merge'].isin(['old', 'conflict'])].copy()
+
+        if not r_chunks1.empty:
+            chunks = r_chunks1.to_dict('records')
+
+            ## Download old chunks for comparisons
+            remote2 = copy.deepcopy(remote1)
+            _ = remote2.pop('version')
+            remote2['cache'] = pathlib.Path(self.previous_path)
+
+            with concurrent.futures.ThreadPoolExecutor(max_workers=threads) as executor:
+                futures = []
+                for chunk in chunks:
+                    remote2['chunk'] = chunk
+                    f = executor.submit(download_results, **remote2)
+                    futures.append(f)
+                runs = concurrent.futures.wait(futures)
+
+            chunks1 = [r.result()['chunk'] for r in runs[0]]
+
+            ## Update the results files
+            if max_workers <= 1:
+                updated_paths = []
+                for chunk in chunks1:
+                    p1 = tu.processing.update_compare_results(chunk, self.max_version_date_key, self.preprocessed_path)
+                    updated_paths.append(p1)
+            else:
+                with concurrent.futures.ProcessPoolExecutor(max_workers=max_workers, mp_context=mp.get_context("spawn")) as executor:
+                    futures = []
+                    for chunk in chunks1:
+                        f = executor.submit(tu.processing.update_compare_results, chunk, self.max_version_date_key, self.preprocessed_path)
+                        futures.append(f)
+                    runs = concurrent.futures.wait(futures)
+
+                updated_paths = [r.result() for r in runs[0]]
+        else:
+            updated_paths = []
+
+        ## Combine paths with original paths
+        combo_paths = list(set(results_paths).union(set(updated_paths)))
+
+        ## diagnostic log
+        self.diagnostics['update_conflicting_results'] = {'pass': True}
+
+        return combo_paths
+
+
+    def update_stations(self, results_paths, threads=60, max_workers=4):
+        """
+
+        """
+        ## Organise the paths
+        paths_dict = {}
+        for path in results_paths:
+            path1 = os.path.split(path)[1]
+            ds_id, version_date1, stn_id, chunk_id, _, _ = path1.split('_')
+
+            if ds_id in paths_dict:
+                if stn_id in paths_dict[ds_id]:
+                    paths_dict[ds_id][stn_id]['results_paths'].append(path)
+                else:
+                    paths_dict[ds_id].update({stn_id: {'results_paths': [path], 'new_path': self.preprocessed_path}})
+            else:
+                paths_dict[ds_id] = {stn_id: {'results_paths': [path], 'new_path': self.preprocessed_path}}
+
+        ## Get the old stations
+        if isinstance(self.public_url, str):
+            remote1 = {'bucket': self.bucket, 'public_url': self.public_url, 'version': self.system_version}
+        else:
+            remote1 = {'bucket': self.bucket, 'connection_config': self.connection_config, 'version': self.system_version}
+
+        t1 = Tethys([remote1])
+
+        dataset_ids = [ds_id for ds_id in paths_dict]
+
+        for ds_id in dataset_ids:
+            if ds_id in t1._datasets:
+                _ = t1.get_stations(ds_id)
+
+        old_stn_dict = t1._stations
+
+        if len(old_stn_dict) > 0:
+            remote1['system_version'] = remote1.pop('version')
+
+            with concurrent.futures.ThreadPoolExecutor(max_workers=threads) as executor:
+                futures = []
+                for ds_id, stn_path in paths_dict.items():
+                    if ds_id in old_stn_dict:
+                        for stn_id in stn_path:
+                            if stn_id in old_stn_dict[ds_id]:
+                                remote1['dataset_id'] = ds_id
+                                remote1['station_id'] = stn_id
+                                f = executor.submit(tu.s3.get_remote_station, **remote1)
+                                futures.append(f)
+                runs = concurrent.futures.wait(futures)
+
+            old_stns = [r.result() for r in runs[0]]
+
+            for stn in old_stns:
+                ds_id = stn['dataset_id']
+                stn_id = stn['station_id']
+                paths_dict[ds_id][stn_id]['old_stn_data'] = stn
+
+        ## Update station data
+        if max_workers <= 1:
+            stn_paths = []
+            for ds_id, stn in paths_dict.items():
+                for stn_id, d in stn.items():
+                    f = tu.processing.update_station(**d)
+                    stn_paths.append(f)
+        else:
+            with concurrent.futures.ProcessPoolExecutor(max_workers=max_workers, mp_context=mp.get_context("spawn")) as executor:
+                futures = []
+                for ds_id, stn in paths_dict.items():
+                    for stn_id, d in stn.items():
+                        f = executor.submit(tu.processing.update_station, **d)
+                        futures.append(f)
+                runs = concurrent.futures.wait(futures)
+
+            stn_paths = [r.result() for r in runs[0]]
+
+        ## diagnostic log
+        self.diagnostics['update_stations'] = {'pass': True}
+
+        return stn_paths
+
+
+    def upload_final_results(self, results_paths, threads=60):
+        """
+
+        """
+        tu.misc.diagnostic_check(self.diagnostics, 'save_new_results')
+        tu.misc.diagnostic_check(self.diagnostics, 'update_stations')
 
         ## Determine which chunks need to uploaded
         if isinstance(self.public_url, str):
@@ -1144,49 +1419,85 @@ class Titan(object):
 
         r_chunks = tu.s3.determine_results_chunks_diffs(results_paths, remote1)
 
+        stn_keys = []
+        results_keys = []
+
         if not r_chunks.empty:
-            results_paths1 = r_chunks['file_path'].tolist()
+            # old_chunks = r_chunks[r_chunks['_merge'] == 'old']
 
-            random.shuffle(results_paths1)
+            # if not old_chunks.empty:
+            #     ## Direct copy S3 objects
+            #     print('Old objects will be copied over to the new version.')
 
-            ## Make sure all of the station files exist
-            stn_paths = []
-            for path in results_paths1:
-                ds_id, version_date1, stn_id, chunk_id, hash_id, _ = os.path.split(path)[1].split('_')
+            #     s3 = s3_client(self.connection_config, max_pool_connections=30)
 
-                stn_file_name = tu.processing.ds_stn_json_str.format(ds_id=ds_id, stn_id=stn_id)
-                stn_file_path = os.path.join(os.path.split(path)[0], stn_file_name)
+            #     old_chunks_list = old_chunks.to_dict('records')
 
-                if os.path.exists(stn_file_path):
-                    stn_paths.append(stn_file_path)
-                else:
-                    raise FileNotFoundError('station file not found: ' + stn_file_path)
+            #     source_dest_keys = []
+            #     for chunk in old_chunks_list:
+            #         version_date_key = tu.misc.make_run_date_key(chunk['version_date'])
+            #         dest_key = tdm.utils.key_patterns[self.system_version]['results'].format(dataset_id=chunk['dataset_id'], version_date=version_date_key, station_id=chunk['station_id'], chunk_id=chunk['chunk_id'])
 
-            ## Iterate through the results and stations
-            with concurrent.futures.ThreadPoolExecutor(max_workers=threads) as executor:
-                s3 = s3_client(self.connection_config, max_pool_connections=threads)
-                futures = []
+            #         if dest_key != chunk['key']:
+            #             source_dest_keys.append({'source_key': chunk['key'], 'dest_key': dest_key})
 
+            #     if len(source_dest_keys) > 0:
+            #         _ = tu.s3.multi_copy_object_s3(s3, self.bucket, self.bucket, source_dest_keys)
+
+            if 'old' in r_chunks._merge.unique():
+                raise ValueError('Looks like you need to run the update_conflicting_results method.')
+
+            other_chunks = r_chunks[r_chunks['_merge'].isin(['conflict', 'new'])]
+
+            if not other_chunks.empty:
+
+                merge_types = other_chunks['_merge'].unique()
+
+                if 'conflict' in merge_types:
+                    tu.misc.diagnostic_check(self.diagnostics, 'update_conflicting_results')
+
+                results_paths1 = other_chunks['file_path'].tolist()
+
+                random.shuffle(results_paths1)
+
+                ## Make sure all of the station files exist
+                stn_paths = []
                 for path in results_paths1:
-                    f = executor.submit(tu.s3.put_result, s3, self.bucket, path, system_version=self.system_version)
-                    futures.append(f)
+                    ds_id, version_date1, stn_id, chunk_id, _, _ = os.path.split(path)[1].split('_')
 
-                for path in set(stn_paths):
-                    # Add the station file
-                    f = executor.submit(tu.s3.put_station, s3, self.bucket, path, system_version=self.system_version)
-                    futures.append(f)
+                    stn_file_name = tu.processing.ds_stn_json_str.format(ds_id=ds_id, stn_id=stn_id)
+                    stn_file_path = os.path.join(os.path.split(path)[0], stn_file_name)
 
-                runs = concurrent.futures.wait(futures)
+                    if os.path.exists(stn_file_path):
+                        stn_paths.append(stn_file_path)
+                    else:
+                        raise FileNotFoundError('station file not found: ' + stn_file_path)
 
-            keys = [r.result() for r in runs[0]]
-            keys.sort()
+                ## Iterate through the results and stations
+                with concurrent.futures.ThreadPoolExecutor(max_workers=threads) as executor:
+                    s3 = s3_client(self.connection_config, max_pool_connections=threads)
+                    futures = []
 
-            stn_keys = [key for key in keys if 'station' in key]
-            results_keys = [key for key in keys if 'results' in key]
+                    for path in results_paths1:
+                        f = executor.submit(tu.s3.put_result, s3, self.bucket, path, system_version=self.system_version)
+                        futures.append(f)
+
+                    for path in set(stn_paths):
+                        # Add the station file
+                        f = executor.submit(tu.s3.put_station, s3, self.bucket, path, system_version=self.system_version)
+                        futures.append(f)
+
+                    runs = concurrent.futures.wait(futures)
+
+                keys = [r.result() for r in runs[0]]
+                keys.sort()
+
+                stn_keys = [key for key in keys if 'station' in key]
+                results_keys = [key for key in keys if 'results' in key]
+            else:
+                print('No chunks have changed or have been added, so no files have been uploaded.')
         else:
-            stn_keys = []
-            results_keys = []
-            print('No chunks have changed, so no files have been uploaded.')
+            print('No chunks have changed or have been added, so no files have been uploaded.')
 
         ## Remove source files
         path1 = pathlib.Path(os.path.split(results_paths[0])[0])
@@ -1214,12 +1525,12 @@ class Titan(object):
             stns_list = tu.s3.put_remote_agg_stations(dataset_id, self.bucket, s3, self.public_url, threads=threads, system_version=self.system_version)
 
             if stns_list:
-                print('-- Updating versions')
+                # print('-- Updating versions')
                 version_data1 = self.version_dict[dataset_id]
 
                 version_list =  tu.s3.update_versions(version_data1, self.bucket, self.connection_config, self.public_url, system_version=self.system_version)
 
-                print('-- Updating Dataset')
+                # print('-- Updating Dataset')
                 ds_new = tu.s3.put_dataset(ds, self.bucket, s3=s3, public_url=self.public_url, system_version=self.system_version)
 
         # Aggregate all datasets for the bucket
